/***********************************************************

File Name :
        ew2msring.C

Programmer:
	Paul Friberg
        original Author: Phil Maechling

Description:
        Earthworm to MiniSeed
        This is an earthworm client which is sits on a Tracebuf ring
        and input tracebuf packets. It then repackages them up into
        MiniSeed packets and put them into an ring.

Creation Date:
        9 March 1999


Usage Notes:
	There is logic here to make the tranmission less bursty. Basically
	I don't want to send all packets every 10 seconds or so, then
	not send anything for 10 seconds. I want to play out the packets
	over the full 10 seconds. The logic to do that is in the
	start slot and AnalogChannel objects.
	
	paulf - translation of the above: Since this was used for ANALOG
	data all of the data would come in in one fell swoop (a Multiplexed
	Data stream all sampled at the same time)...so Phil instituted a
	spooling scheme that spooled different SCN's based on different
	indexes in the channel config table (file). This may not be necessary
	for other digital stations since they all send at different times and
	have different packetization and delivery latencies...


Modification History:

	12 June 2001 - Paul Friberg - changing the logic to deal with non
		Analog type stations and increase the channel buffer to
		something a bit larger than 43,000 samples which turns out
		to be a number based on the old Analog system.

	14 June 2001 - PF - introduced a versioning number system 

	20 June 2001 - PF - changed startup logic to keep all data
				also changed TracebufsPerSend to SamplesPerSend

	25 June 2001 - PF - added RestartOnGap to serve analog stations

	19 June 2002 - PF - added check to scan_netchan_config()

	11 Sept 2002 - PF - changed logit_init to use config file as per
				new version of EW

	22 Dec 2003 - PF - 1.0.6 - modified the program to throw away any
				packets that are more than 1 hour in the
				future (bogus timing packet observed from
				EW hub wreaked havoc with wavepools).

	29 Dec 2003 - PF - 1.0.7 - added in heartbeat changes from Jan 2003:
			 moved heartbeat to  right after the tport_getmsg() call
 
                Moved the heartbeat processing there because then heartbeats will
                be issued, even when data input to the ring stops....
                Previously the statmgr thought that the program was dead
                because heartbeats stopped, yet it was just the input data
                that had stopped.       

	30 Dec 2003 - PF - 1.0.8 - moved the get_my_wordorder() out of the 
		Compressor::MakePackets() call and into Compressor() constructor.
		Not having this in that location was causing a problem with the
		first packet through the compressor which essentially made the
		channel have bogus data.

	5 Jan 2004 - PF - 1.0.9 - changed the default check on future packets 
		to be 10 seconds from the end time (not the  1 hour ahead of start time)

	5 Jan 2004 - PF - 1.0.10 - added in an error message to be sent to statmgr if
		any future timestamped packets are observed. This has  a limited throttle
		so as to not flood any mailbox if the timestamps get out of whack.
	
	7 Jan 2004 - PF - 1.0.11 - fixed the timestamp on the statmgr error message
		from an incorrect value of 0 (unix Jan 1, 00:00 1970) to  the correct
		time.

	16 Aug 2004 - PNL - 1.1.0 - Merged UCB's ew2m with CIT's ew2mcast.
		Added NSlots config directive to control staggered start.
		Added SEED renaming.
		Cleaned up lots of old junk based on Earthworm mythology.
		Removed last of a2m and analog names.
		Added support for SEED location codes in Earthworm.

        2005/11/13 (2005.318) - 1.00 - DSN
		Derived initial ew2orbms from ew2mcast.

	2010/12/28 (2010.362) - 1.01 - DSN
		Derived initial ew2msring from ew2msorb.

	2012/04/16 (2012.107) - 1.02 - DSN
		Fixed bug in pressor.C that was not freeing MiniSEED record on SendMsg() error.

	2013/06/24 (2013.175) - 1.03 - DSN
		Exit on error writing to ringserver. Assume someone will restart it.

	2013/07/15 (2013.196) - 1.04 - DSN
		Include PID in heartbeat messages.
		Skip data too large for STEIM2 compression.
**********************************************************/

#define VER_NO "1.04 (2013.196)"
extern "C" {
    char *version = VER_NO;
}

// list the error messages here (only 1 for now)
#define EW2MSRING_ERR_FUTUREDATA 0

#include <iostream>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

extern "C" {
#include <earthworm.h>
#include <transport.h>
#include <trace_buf.h>
#include <swap.h>
}
#include <map>
#include "ew2msring.h"
#include "ew2msring_config.h"
#include "netchancfg.h"
#include "RetCodes.h"
#include "msringwriter.h"
#include "Channel.h"
#include "TimeStamp.h"
#include "ewchan.h"
#include "seismic.h"


#define ew2m_debug TN_FALSE

#define FUTURE_THROTTLE 10 // after this many future errs, wait 20 min before sending more

typedef std::map<Channel,EWChannel,std::less<Channel>, std::allocator<std::pair<const Channel, EWChannel> > > NetworkMap;

NetworkMap chanmap;
char *cmdname;

/************************************************************************/
/*  finish_handler:							*/
/*	Signal handler -  sets terminate flag so we can exit cleanly.	*/
/************************************************************************/

static int terminate_proc = 0;
extern "C" void finish_handler(int sig)
{
    signal (sig,finish_handler);    /* Re-install handler (for SVR4)	*/
    terminate_proc = 1;
    return;
}

int main( int argc, char **argv )
{
   int           i;               /* Loop counter */
   char          *WaveBuf;        /* Pointer to waveform buffer */
   TRACE2_HEADER  *WaveHead;       /* Pointer to waveform header */
   long          *WaveLong;       /* Long pointer to waveform data */
   short         *WaveShort;      /* Short pointer to waveform data */
   long          MsgLen;          /* Size of retrieved message */
   MSG_LOGO      getlogo;         /* Logo of requested waveforms */
   MSG_LOGO      logo;            /* Logo of retrieved msg */
   MSG_LOGO      hrtlogo;         /* Logo of outgoing heartbeats */
   MSG_LOGO      errlogo;         /* Logo of outgoing error messages to statmgr */
   int           Nsta;            /* Number of stations in list */
   time_t        then;            /* Previous heartbeat time */
   long          InBufl;          /* Maximum message size in bytes */
   GPARM         Gparm;           /* Configuration file parameters */
   EWH           Ewh;             /* Parameters from earthworm.h */
   double  	 d_curtime;	/* used for future timestamp error checking */
   char 	 err_msg[256]; 	/* where we put the error message before sending it to statmgr */
   int		 future_err_counter; /* a throtle on the number of future error messages sent */
   double	 future_time;	/* holds the time of the last future err message sent */
   char          *configfile;      /* Pointer to name of config file */
   pid_t         myPid;            /* Process id of this process */

   future_err_counter = 0;
   future_time  = 0.0;

   init_qlib2(1); //Added by Kalpesh.
   set_hdr_wordorder(SEED_BIG_ENDIAN);
   set_data_wordorder (SEED_BIG_ENDIAN);

/* Catch broken socket signals
   ******************************/
   (void)sigignore(SIGPIPE);

   terminate_proc = 0;
   signal (SIGINT,finish_handler);
   signal (SIGTERM,finish_handler);

/* Check command line arguments
   ****************************/
   cmdname = tail(argv[0]);
   fprintf (stdout, "%s version %s\n", cmdname, version);
    
   if ( argc != 2 )
   {
      fprintf( stderr, "Usage: %s <configfile>\n", cmdname );
      return -1;
   }
   configfile = argv[1];

/* Initialize name of log-file & open it
   *************************************/
   logit_init( configfile, 0, 256, 1 );

/* Get parameters from the configuration files
   *******************************************/
   if ( GetConfig( argv[1], &Gparm ) == -1 )
   {
      fprintf( stderr, "%s: GetConfig() failed. Exiting.\n", cmdname );
      return -1;
   }

/* Look up info in the earthworm.h tables
   **************************************/
   if ( GetEwh( &Ewh ) < 0 )
   {
      fprintf( stderr, "%s: GetEwh() failed. Exiting.\n", cmdname );
      return -1;
   }

/* Specify logos of incoming waveforms and outgoing heartbeats
   ***********************************************************/
   getlogo.instid = Ewh.GetThisInstId;
   getlogo.mod    = Ewh.GetThisModId;
   getlogo.type   = Ewh.TypeWaveform;

   hrtlogo.instid = Ewh.MyInstId;
   hrtlogo.mod    = Gparm.MyModId;
   hrtlogo.type   = Ewh.TypeHeartBeat;

   errlogo.instid = Ewh.MyInstId;
   errlogo.mod    = Gparm.MyModId;
   errlogo.type   = Ewh.TypeError;

/* Get our own pid for restart purposes
   ************************************/
   myPid = getpid();
   if ( myPid == -1 )
   {
      logit( "e", "%s: Can't get my pid. Exiting.\n", cmdname );
      return -1;
   }

/* Log the configuration parameters
   ********************************/
   LogConfig( &Gparm );

/* Allocate the waveform buffer
   ****************************/
/* This calculation is 2 * MAX_TRACEBUF_SIZ, to account
   for the fact that we may convert from shorts into ints.
   But of course MAX_TRACEBUF_SIZ says how many _bytes_ could in a TRACEBUF 
   messages, not how many samples. Oh well, this is big enough. */

   InBufl = MAX_TRACEBUF_SIZ * 2;

   WaveBuf = (char *) malloc( (size_t) InBufl );
   if ( WaveBuf == NULL )
   {
      logit( "et", "%s: Cannot allocate waveform buffer\n", cmdname );
      return -1;
   }

/* Point to header and data portions of waveform message
   *****************************************************/

   WaveHead  = (TRACE2_HEADER *) WaveBuf;
   WaveLong  = (long *) (WaveBuf + sizeof(TRACE2_HEADER));
   WaveShort = (short *) (WaveBuf + sizeof(TRACE2_HEADER));

/* Read in the Channel List which identifies the channels which
   will be formatted and output.
 **********************************************************/

  struct channel_identifier_type channels[MAX_CHANNELS_IN_NETWORK];

  int number_of_channels_found = 0;

  int res = scan_netchannel_config_file(Gparm.ChannelFile,
                                 channels,
                                 number_of_channels_found,
                                 TN_FALSE);

  if (res != TN_SUCCESS)
  {
     std::cout << "Error reading channel selection configuration file " 
          << Gparm.ChannelFile << std::endl;
     std::cout << "Exiting " << argv[0] << std::endl;
     return(TN_FAILURE);
  } else {
     std::cout << "Found " << number_of_channels_found << 
	  " channels in " << Gparm.ChannelFile << std::endl;
  }

/* Find Samples per Send from Configuration Information */

  int samples_per_send = Gparm.SamplesPerSend;

  std::cout << "Samples per Send is : " << samples_per_send << std::endl;

/* Create the EWChannel Map 
   *******************************************************/

  std::pair< NetworkMap::iterator, bool> mpair;
  Channel ewchan;
  Channel seedchan;
  EWChannel* ewchannel;
  int channel_count = 0;

  for(int ix=0;ix<number_of_channels_found;ix++)
  {
        strcpy(ewchan.network,channels[ix].network);
        strcpy(ewchan.station,channels[ix].station);
        strcpy(ewchan.channel,channels[ix].channel);
        strcpy(ewchan.location,channels[ix].location);

        strcpy(seedchan.network,channels[ix].seednetwork);
        strcpy(seedchan.station,channels[ix].seedstation);
        strcpy(seedchan.channel,channels[ix].seedchannel);
        strcpy(seedchan.location,channels[ix].seedlocation);

	int startslot = channel_count % Gparm.NSlots;

        std::cout << "SNCL in channel.cfg found in station file " << 
          channels[ix].station << "." <<
          channels[ix].network << "." << 
          channels[ix].channel << "." <<
	  channels[ix].location << " " <<
          channels[ix].seedstation << "." <<
          channels[ix].seednetwork << "." << 
          channels[ix].seedchannel << "." <<
	  channels[ix].seedlocation << " " <<
	  startslot << std::endl;

        ewchannel = new EWChannel(ewchan,
				     seedchan,
				     samples_per_send,
				     startslot,
				     Gparm.CompressionLevel);

        mpair = chanmap.insert(NetworkMap::value_type(ewchan,*ewchannel) );

        delete ewchannel;

        channel_count++;

        if( mpair.second == TN_FALSE)
        {
          std::cout << "Error (" << cmdname << ") : Duplicate Channels in channel Map"
           << std::endl;
      
          std::cout << "Duplicate Channels in channel.cfg " << 
            channels[ix].station << "." <<
            channels[ix].network << "." << 
            channels[ix].channel << "." <<
	    channels[ix].location <<std:: endl;

          return(TN_FAILURE);
        } 
  }

  int number_of_achans = chanmap.size();

  if(number_of_achans > 0)
  {
    std::cout << "Number of EW Channels to convert to MiniSEED " <<
	number_of_achans << std::endl;
  }
  else
  {
    std::cout << "Zero Channels added to Map. Exiting" << std::endl;
    return(TN_FAILURE);
  }

/* Attach to existing transport rings
   **********************************/

   tport_attach( &Gparm.InRegion, Gparm.InKey );

/* Create and initializa an RingWriter object.
   *******************************************************/
   MSRingWriter ring(Gparm.RingServerName,cmdname);
   res = ring.InitializeRing();
   std::cout << "Opened ring " << Gparm.RingServerName << std::endl;

/* Flush the input earthworm ring
   ********************/

   while ( tport_getmsg( &Gparm.InRegion, &getlogo, 1, &logo, &MsgLen,
                          WaveBuf, MAX_TRACEBUF_SIZ) == GET_OK );

/* Get the time when we start reading messages.
   This is for issuing heartbeats.
   *******************************************/
   time( &then );
   // Make it so we issue the first heartbeat right away
   then -= Gparm.HeartbeatInt;

   int sendrc = TN_SUCCESS;

/* Loop to read waveform messages and invoke the MSEED Builder
   ****************************************************/
   while ( tport_getflag( &Gparm.InRegion ) != TERMINATE  &&
           tport_getflag( &Gparm.InRegion ) != myPid && (! terminate_proc))
   {
      char    type[3];
      int     rc;               /* Return code from tport_getmsg() */
      time_t  now;              /* Current time */
      double  GapSizeD;         /* Number of missing samples (double) */
      long    GapSize;          /* Number of missing samples (integer) */


      /* 
	 Moved the heartbeat processing here because then heartbeats will
	 be issued, even when data input to the ring stops....
	 previously the statmgr thought that the program was dead
	 because heartbeats stopped, yet it was just the input data
	 that stopped.
	 Add a test to see if it is
	 time for a heartbeat. If so, send a heartbeat to the transport ring.
      ********************************************************************/

      time( &now );
      if ( Gparm.HeartbeatInt > 0 && (now - then) >= Gparm.HeartbeatInt )
      {
         int  lineLen;
         char line[40];

         then = now;

	 sprintf( line, "%ld %ld\n", (long) now, (long) myPid );
         lineLen = strlen( line );

         if ( tport_putmsg( &Gparm.InRegion, &hrtlogo, lineLen, line ) !=
              PUT_OK )
         {
            logit( "et", "%s: Error sending heartbeat. Exiting.", cmdname );
            break;
         }
      }

/* Get a waveform buffer from transport region
   *******************************************/

      rc = tport_getmsg( &Gparm.InRegion, &getlogo, 1, &logo, &MsgLen,
                         WaveBuf, MAX_TRACEBUF_SIZ);

      if ( rc == GET_NONE )
      {
         sleep_ew( 100 );
         continue;
      }

      if ( rc == GET_NOTRACK )
         logit( "et", "%s: Tracking error.\n", cmdname );

      if ( rc == GET_MISS_LAPPED )
         logit( "et", "%s: Got lapped on the ring.\n", cmdname );

      if ( rc == GET_MISS_SEQGAP )
         logit( "et", "%s: Gap in sequence numbers.\n", cmdname );

      if ( rc == GET_MISS )
         logit( "et", "%s: Missed messages.\n", cmdname );

      if ( rc == GET_TOOBIG )
      {
         logit( "et", "%s: Retrieved message too big (%ld) for msg.\n", cmdname,
                MsgLen );
         continue;
      }

/* If necessary, swap bytes in the message
   ***************************************/

      if ( WaveMsg2MakeLocal( WaveHead ) < 0 )
      {
         logit( "et", "%s: Unknown waveform type.\n", cmdname );
         continue;
      }

/* new in 2003 - check for timestamps somewhere in the future and
	reject any data that have timing greater than X secs ahead
	system time. Currently hardcoded at 10 seconds!
 */
    d_curtime = (double) time(0);
    if ((WaveHead->endtime-d_curtime) > Gparm.future_data_threshold) {
         logit( "et", "%s: %s.%s.%s.%s data timestamped %f secs in future.\n",
		cmdname, WaveHead->net, WaveHead->sta, WaveHead->chan,
		WaveHead->loc, (WaveHead->endtime-d_curtime) );

	// the following snippet of code throttles the number of error messages sent
	 if (future_time != 0.0 && future_err_counter > FUTURE_THROTTLE) {
		if ((d_curtime - future_time) >= 1200) {
		    // if after 10 future error messages and  20 min
		    // have elapsed since the same error, restart counter
		    future_err_counter = 0;
		} else {
		    continue; // don't send an error message till time has elapsed
		}
	 }
	 time_t msg_time;
 	 time(&msg_time);
	 future_err_counter++;
	 future_time = d_curtime;

	 sprintf( err_msg, "%ld %hd %s.%s.%s.%s had future timestamped data of %f by %f seconds\n", 
		  msg_time, EW2MSRING_ERR_FUTUREDATA, 
		  WaveHead->net, WaveHead->sta, WaveHead->chan, WaveHead->loc,
		  WaveHead->endtime, (WaveHead->endtime-d_curtime) );

         if ( tport_putmsg( &Gparm.InRegion, &errlogo, strlen(err_msg), err_msg ) !=
              PUT_OK )
         {
            logit( "et", "%s: Error sending error message about future problem.", cmdname );
         }
         continue;
    }

/* If the samples are shorts, make them longs
   ******************************************/
      strcpy( type, WaveHead->datatype );

      if ( (strcmp(type,"i2")==0) || (strcmp(type,"s2")==0) )
      {
         for ( i = WaveHead->nsamp - 1; i > -1; i-- )
            WaveLong[i] = (long) WaveShort[i];
      }

/* Here we have a TraceBuf Packet in the Wavelong array. We now create
   a packet object, then try to add the packet into the channel objects.
  **********************************************************************/

      Channel thispacket;

      strcpy(thispacket.network,WaveHead->net);
      strcpy(thispacket.station,WaveHead->sta);
      strcpy(thispacket.channel,WaveHead->chan);
      // TRACEBUF packets always use "--" instead of empty location code
      if (WaveHead->loc[0] == '-')
	  thispacket.location[0] = '\0';
      else
	  strcpy(thispacket.location,WaveHead->loc);

      
      NetworkMap::iterator p = chanmap.find(thispacket);

      if(p != chanmap.end() )
      {
        if(ew2m_debug)
        {
           std::cout << "Received Data for this Channel " <<
             (*p).second.ewchannel.station << "." << 
             (*p).second.ewchannel.network << "." << 
             (*p).second.ewchannel.channel << "." <<
             (*p).second.ewchannel.location << "." << std::endl;
        }

        //
        // Do the staggered start testing here. Once a station is started,
        // ignore the start testing. But here at the start, test each channel
        // to see if the start slot number matches the sequence number.
        // When the packet sequence number first matches the startslot, then
        // we explicitly flush the data channel.
	// The startslot is based on the NSlots.
	// However, since each channel (or even each tracebuf for each channel)
	// may have a different number of samples, this is not guaranteed to
	// distribute the network traffic.
        //

	//
        // First we test for timegaps. If there is a timegap, we
        // send all the data we current have buffered up. Then, we
        // reset the channel to its initial state.
        //

	int res = (*p).second.TimeGap(WaveHead);

	if(res == TN_TRUE)
	{
	  TimeStamp th(EARTHWORM_TIME, (*p).second.TimeOf(WaveHead));
	  TimeStamp te(EARTHWORM_TIME, (*p).second.TimeLastEnd());
	  std::cout << "Found Time Gap for " << 
             (*p).second.ewchannel.station << "." << 
             (*p).second.ewchannel.network << "." << 
             (*p).second.ewchannel.channel << "." <<
             (*p).second.ewchannel.location << "." << " - ";
	  std::cout << "Found " << th << " Last " << te << std::endl;

	  //
	  // Process a time gap here.
	  //

	  if(!(*p).second.BufferIsEmpty())
	  {
	    sendrc = (*p).second.SendPacket(ring);
	  }
	  if (Gparm.RestartOnGap) 
	      (*p).second.ResetToNotStarted();
	}

	if (! (*p).second.BufferHasRoom(WaveHead->nsamp)) 
	{
	  if(!(*p).second.BufferIsEmpty())
	  {
	    sendrc = (*p).second.SendPacket(ring);
	  }
	}

	// Add the data from this tracebuf to the channel buffer.
	if ((*p).second.BufferHasRoom(WaveHead->nsamp))
	{
	  (*p).second.AddDataToBuffer(WaveBuf);
	}
	else {
	  // Packet must be longer than our channel buffer.
	  // Complain and discard it.
          logit( "et", "%s: Wavebuf too long - %d samples. Discarding\n", 
		cmdname, (int)(WaveHead->nsamp) );
	}

	// Check to see if we have started this channel's slot.
	int juststarted = 0;
        if(!(*p).second.Started())
        {
          int seqno = (*p).second.NextSequenceNumber() %
				Gparm.NSlots;
          if (seqno == (*p).second.StartSlot() )
	  {
            (*p).second.SetStarted();
	    juststarted = 1;
	  }
	}
	if((*p).second.BufferIsFull() || juststarted)
	{
	  sendrc = (*p).second.SendPacket(ring);
	}
      }
      else // This is  the path when we get a packet which is not in the map.
      {
	if(ew2m_debug)
	{
	  std::cout << "Channel not in NetMap " <<
	    thispacket.station << "." <<
	    thispacket.network << "." <<
	    thispacket.channel << "." <<
	    thispacket.location << std::endl;
	}      
      }
      if (sendrc != TN_SUCCESS) terminate_proc = 1;
   }

/* Detach from the ring buffers
   ****************************/
  
   tport_detach( &Gparm.InRegion );
   logit( "t", "%s: Termination requested\n.", cmdname );

/* Iterate over all channels, flushing any buffered data. */
   NetworkMap::iterator p = chanmap.begin();
   while (p != chanmap.end()) 
   {
     if(! (*p).second.BufferIsEmpty())
	{
	  (*p).second.SendPacket(ring);
	}
     p++;
   }

   logit( "t", "%s: Exiting\n.", cmdname );
   return 0;

}

      /*******************************************************
       *                      GetEwh()                       *
       *                                                     *
       *      Get parameters from the earthworm.h file.      *
       *******************************************************/

int GetEwh( EWH *Ewh )
{
   if ( GetLocalInst( &Ewh->MyInstId ) != 0 )
   {
      fprintf( stderr, "%s: Error getting MyInstId.\n", cmdname );
      return -1;
   }

   if ( GetInst( "INST_WILDCARD", &Ewh->GetThisInstId ) != 0 )
   {
      fprintf( stderr, "%s: Error getting GetThisInstId.\n", cmdname );
      return -2;
   }
   if ( GetModId( "MOD_WILDCARD", &Ewh->GetThisModId ) != 0 )
   {
      fprintf( stderr, "%s: Error getting GetThisModId.\n", cmdname );
      return -3;
   }
   if ( GetType( "TYPE_HEARTBEAT", &Ewh->TypeHeartBeat ) != 0 )
   {
      fprintf( stderr, "%s: Error getting TypeHeartbeat.\n", cmdname );
      return -4;
   }
   if ( GetType( "TYPE_ERROR", &Ewh->TypeError ) != 0 )
   {
      fprintf( stderr, "%s: Error getting TypeError.\n", cmdname );
      return -5;
   }
   if ( GetType( "TYPE_TRACEBUF2", &Ewh->TypeWaveform ) != 0 )
   {
      fprintf( stderr, "%s: Error getting TYPE_TRACEBUF2.\n", cmdname );
      return -8;
   }
   return 0;
}
