/***********************************************************

File Name :
        ewchan.C

Programmer:
        Phil Maechling

Description:
        EW Channel Class
        This is a class definition which encapsulates an EW channel.
        The hope is to create a series of these, then put them in a map,
	and when we get an EW packet, we add it to the channel
        object, and when we have enough data, we send the data out.

Creation Date:
        13 March 1999


Usage Notes:
	There is an assumption the the user of an EW channel
	will send data immediately after detecting a full buffer.
	If they receive another packet, then send the data, they
	buffer will overflow.


Modification History:

16 August 2004 - added initialization/reset of p_last_sample


**********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include "msringwriter.h"
#include "Channel.h"
#include "RetCodes.h"
#include "ewchan.h"
#include "trace_buf.h"
#include "seismic.h"


const int ewchan_debug = TN_FALSE;

EWChannel:: EWChannel(const Channel         ewchan1,
		      const Channel	    seedchan1,
		      const int		    samples_per_send,
		      const int		    start_slot,
		      const int             compression_level)
{
  ewchannel           = ewchan1;
  seedchannel	      = seedchan1;
  p_samples_per_send  = samples_per_send;
  p_start_slot        = start_slot;


  if(compression_level == 1)
  {
    p_compression_level = STEIM1;
  }
  else if(compression_level == 2)
  {
    p_compression_level = STEIM2;
  } 
  else
  {
    std::cout << "Invalid Compression Level " << compression_level 
     << " : Setting it to 1." << std::endl;
    p_compression_level = STEIM1;
  }

  p_lastendtime  = 0.0;
  p_last_sample = 0;
  p_buffer_is_empty = TN_TRUE;
  p_buffer_is_full  = TN_FALSE;
  p_number_of_samples = 0;
  p_gap_found = TN_FALSE;
  trace_hdr.nsamp = 0;
  p_started = TN_FALSE;
  p_sequence_number = 0;

  for(int i = 0; i< MAX_SAMPLES_IN_CHANNEL_BUFFER;i++)
  {
    diff_array[i] =0;
  }
}

void EWChannel::SetStarted()
{
  p_started = TN_TRUE;
}

void EWChannel::SetNotStarted()
{
  p_started = TN_FALSE;
}


int EWChannel::Started()
{
  return(p_started);
}

int EWChannel::NextSequenceNumber()
{

  // I'm noting that this sequence number is used only by ewchan to
  // determine the startslot of a channel. The pressor routine
  // carries it's own sequence number which it puts into the packets.
  // I could pass this to the pressor routine if it makes a difference.
  //

  p_sequence_number++;
  p_sequence_number = p_sequence_number % 1000000;
  return(p_sequence_number);
}

int EWChannel::StartSlot()
{
  return(p_start_slot);
}

EWChannel:: ~EWChannel()
{

}

int EWChannel::TimeGap(TRACE2_HEADER *header)
{

  double GapSizeD;
  long   GapSize;

  if (p_lastendtime == 0.0) return (TN_FALSE);

  GapSizeD = header->samprate * (header->starttime - p_lastendtime);
 
  if(GapSizeD < 0.0)
  {
    std::cout << "Backward time step found in EW data" << std::endl;
    p_gap_found = TN_TRUE;
    return(TN_TRUE);
  } 
  else
  {
    GapSize = (long) (GapSizeD + 0.5);
  }
    
  if (GapSize > 1)
  {
    p_gap_found = TN_TRUE;
    return(TN_TRUE);
  }
  else
  {
    return(TN_FALSE);
  }
}

int EWChannel::BufferIsFull()
{
  return(p_buffer_is_full);
}


int EWChannel::BufferIsEmpty()
{
 return(p_buffer_is_empty);
}


int EWChannel::BufferHasRoom(int nsamples)
{
  return(p_number_of_samples+nsamples<MAX_SAMPLES_IN_CHANNEL_BUFFER);
}


void EWChannel::AddDataToBuffer(char* buf)
{

  long* data = (long *) (buf + sizeof(TRACE2_HEADER));

  TRACE2_HEADER  temp_hdr;

  if(p_buffer_is_empty)
  {
    memcpy(&temp_hdr,buf,sizeof(TRACE2_HEADER));
    temp_hdr.nsamp = 0;
  }
  else
  {
   temp_hdr = trace_hdr;
  }

  memcpy(&trace_hdr,buf,sizeof(TRACE2_HEADER));

  trace_hdr.starttime = temp_hdr.starttime;

  for(int i = 0;i<trace_hdr.nsamp;i++)
  {
      count_array[p_number_of_samples + i] = data[i];
  }

  trace_hdr.nsamp     = temp_hdr.nsamp + 
				trace_hdr.nsamp;

  p_number_of_samples = trace_hdr.nsamp;

  p_lastendtime = trace_hdr.endtime;

  if(p_number_of_samples >= p_samples_per_send)
  {
      p_buffer_is_full = TN_TRUE;
  }

  p_buffer_is_empty = TN_FALSE;
}


int  EWChannel::SendPacket(MSRingWriter& ring)
{

  //
  // Reset the counting variables
  //

  if(ewchan_debug)
  {
    std::cout << "Sending Packet for SNCL " << 
        seedchannel.station << "." <<
	seedchannel.network << "." <<
        seedchannel.channel << "." << 
        seedchannel.location << "." << 
	trace_hdr.nsamp << " samples." << std::endl;
  }

  //
  // Prepare the difference array
  //

  diff_array[0] = count_array[0] - p_last_sample;
  for(int x=1;x<trace_hdr.nsamp;x++)
  {
    diff_array[x] = count_array[x] - count_array[x-1];
  }

  int res = compress.MakePackets(count_array,
				 diff_array,
		                 &trace_hdr,
		                 trace_hdr.nsamp,
                                 p_compression_level,
				 &seedchannel,
				 ring);


  p_last_sample = count_array[trace_hdr.nsamp -1];
  p_buffer_is_empty = TN_TRUE;
  p_buffer_is_full  = TN_FALSE;
  trace_hdr.nsamp = 0;
  p_number_of_samples = 0;

  return(res);
}


void EWChannel::ResetToNotStarted()
{
  p_lastendtime  = 0.0;
  p_buffer_is_empty = TN_TRUE;
  p_buffer_is_full  = TN_FALSE;
  p_number_of_samples = 0;
  p_gap_found = TN_FALSE;
  trace_hdr.nsamp = 0;
  p_started = TN_FALSE;
  p_sequence_number = 0;

  for(int i = 0; i< MAX_SAMPLES_IN_CHANNEL_BUFFER;i++)
  {
    diff_array[i] =0;
  }
}

double EWChannel::TimeOf(TRACE2_HEADER *header)
{
  return (header->starttime);
}

double EWChannel::TimeLastEnd()
{
  return (p_lastendtime);
}

//
// Copy constructor
//
EWChannel::EWChannel(const EWChannel& c)
{

  ewchannel           = c.ewchannel;
  seedchannel         = c.seedchannel;
  p_samples_per_send  = c.p_samples_per_send;
  p_start_slot        = c.p_start_slot;
  p_compression_level = c.p_compression_level;

  p_lastendtime       = c.p_lastendtime;
  p_buffer_is_empty   = c.p_buffer_is_empty;
  p_buffer_is_full    = c.p_buffer_is_full;
  p_number_of_samples = c.p_number_of_samples;
  p_gap_found         = c.p_gap_found;
  trace_hdr.nsamp     = c.trace_hdr.nsamp;
  p_started           = c.p_started;
  p_sequence_number   = c.p_sequence_number;
  p_last_sample       = c.p_last_sample;

  for(int i = 0; i< MAX_SAMPLES_IN_CHANNEL_BUFFER;i++)
  {
    diff_array[i] =0;
  }
}

EWChannel& EWChannel::operator=(const EWChannel &c)
{
  strcpy(this->ewchannel.network,c.ewchannel.network);
  strcpy(this->ewchannel.station,c.ewchannel.station);
  strcpy(this->ewchannel.channel,c.ewchannel.channel);
  strcpy(this->ewchannel.location,c.ewchannel.location);
  strcpy(this->seedchannel.network,c.seedchannel.network);
  strcpy(this->seedchannel.station,c.seedchannel.station);
  strcpy(this->seedchannel.channel,c.seedchannel.channel);
  strcpy(this->seedchannel.location,c.seedchannel.location);
  this->ewchannel.samprate = c.ewchannel.samprate;
  this->seedchannel.samprate = c.seedchannel.samprate;

  return(*this);

}

int operator<(const EWChannel& ewchan1,
                             const EWChannel& ewchan2)
{
  if (strcmp(ewchan1.ewchannel.station, ewchan2.ewchannel.station) < 0) {
    return(TN_TRUE);
  }
  if ((strcmp(ewchan1.ewchannel.channel, ewchan2.ewchannel.channel) == 0)
    && (strcmp(ewchan1.ewchannel.station, ewchan2.ewchannel.station) < 0)) {
    return(TN_TRUE);
  }

  if ((strcmp(ewchan1.ewchannel.network, ewchan2.ewchannel.network) == 0)
    && (strcmp(ewchan1.ewchannel.station, ewchan2.ewchannel.station) == 0)
    && (strcmp(ewchan1.ewchannel.channel, ewchan2.ewchannel.channel) < 0)) {
    return(TN_TRUE);
  }

  if ((strcmp(ewchan1.ewchannel.network, ewchan2.ewchannel.network) == 0)
    && (strcmp(ewchan1.ewchannel.station, ewchan2.ewchannel.station) == 0)
    && (strcmp(ewchan1.ewchannel.channel, ewchan2.ewchannel.channel) == 0)
    && (strcmp(ewchan1.ewchannel.location, ewchan2.ewchannel.location) == 0)) {
    return(TN_TRUE);
  }

  return(TN_FALSE);
}


int operator==(const EWChannel& ewchan1,
	                      const EWChannel& ewchan2)
{
  
  if ((strcmp(ewchan1.ewchannel.network, ewchan2.ewchannel.network) == 0)
    && (strcmp(ewchan1.ewchannel.station, ewchan2.ewchannel.station) == 0)
    && (strcmp(ewchan1.ewchannel.channel, ewchan2.ewchannel.channel) == 0)
    && (strcmp(ewchan1.ewchannel.location, ewchan2.ewchannel.location) < 0)) {
    return(TN_FALSE);
  }
  return(TN_TRUE);
}
