/***********************************************************

File Name :
        ewchan.h

Programmer:
        Phil Maechling (original), adapted by Doug Neuhauser

Description:
        EW (Earthworm) Channel Class
        This is a class definition which encapsulates an EW channel.
        The hope is to create a series of these, then put them in a map,
	and when we get an EW packet, we add it to the channel
        object, and when we have enough data, we send the data out.

Creation Date:
        13 March 1999, adapted 01 May 2000


Usage Notes:

	There is logic in here (start slot etc) which is supposed to
	allow the system to send out packets spread out over a few
	time windows. Basically, I'm trying to prevent a bursty system.
	The penalty is that I will drop the first set of packets on
	startup and restart until a channels start slot comes up and
	it starts to send.


Modification History:

13 June 2001 - added way to get the gap size
16 August 2004 - added support for earthworm location codes.

**********************************************************/
#ifndef EW_channel_H
#define EW_channel_H


#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "msringwriter.h"
#include "RetCodes.h"
#include "trace_buf.h"
#include "pressor.h"
#include "seismic.h"

// set to 1 minute at 100 Hz sampling
#define MAX_SAMPLES_IN_CHANNEL_BUFFER 6000

class EWChannel
{

  public:

    EWChannel(const Channel	ewchan1,
	      const Channel	seedchan1,
	      const int		samples_per_send,
	      const int		start_slot,
	      const int	        CompressionLevel);

    EWChannel(const EWChannel&);

    ~EWChannel();

    int  TimeGap(TRACE2_HEADER* trace);

    int  BufferIsFull();

    int  BufferIsEmpty();

    int  BufferHasRoom(int nsamples);

    void AddDataToBuffer(char* packet);

    int  SendPacket(MSRingWriter& ring);

    int          Started();
    void         SetStarted();
    void         SetNotStarted();


    void         ResetToNotStarted();

    int          NextSequenceNumber();
    int          StartSlot();

    double	TimeOf(TRACE2_HEADER *header);

    double      TimeLastEnd();
    
    EWChannel& operator=(const EWChannel &c);

    friend int operator<(const EWChannel& ewchan1,
		  const EWChannel& ewchan2);

    friend int operator==(const EWChannel& ewchan1,
		   const EWChannel& ewchan2);


    int             chan_multicast_port;
    Channel         ewchannel;
    Channel	    seedchannel;
    Compressor      compress;

    TRACE2_HEADER trace_hdr;

    int          count_array[MAX_SAMPLES_IN_CHANNEL_BUFFER];
    int          diff_array[MAX_SAMPLES_IN_CHANNEL_BUFFER];

  private:

   double p_lastendtime;
   int    p_buffer_is_empty;
   int    p_buffer_is_full;
   int    p_number_of_samples;
   int    p_gap_found;
   int    p_last_sample;
   int    p_started;
   int    p_sequence_number;
   int    p_start_slot;
   int    p_compression_level;
   int    p_samples_per_send;
};

#endif
