/***********************************************************

File Name :
        msringwriter.C

Programmer:
        Doug Neuhauser (derived from code by Phil Maechling)

Description:
        This is a MiniSEED Ringserver Writer Object.
        It is for writing MiniSEED packets to the ring only, at this point.

Creation Date:
	2005/11/13 (2005.317)

Usage Notes:



Modification History:



**********************************************************/
#include <iostream>
using namespace std;

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <termio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h> 
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "msringwriter.h"
#include "RetCodes.h"

#include <libdali.h>
#include "qlib2.h"

const int MSTB = 0x4D53;      /* MiniSEED packet Ringserver packet type              */
const int MSH  = 0x0D53;      /* MiniSEED packet Ringserver header type              */
const int MAX_MSRECORD_SIZE = 8192;

extern "C" {
extern int ringopen ( char *name, char *perm );
extern int ringclose ( int ring );
extern int ringput ( int ring, char *srcname, double time, char *packet, int nbytes );
}

MSRingWriter::MSRingWriter(const char *ringserver_string, const char *client_id_string)
{
    ring_conn = NULL;
    ringserver = (ringserver_string) ? strdup(ringserver_string) : strdup ("");
    client_id = (client_id_string) ? strdup(client_id_string) : strdup ("");
    ring_conn = dl_newdlcp(ringserver, client_id);
    ack_flag = 0;
    if (ring_conn == NULL) {
	std::cerr << "Error: unable to create dlcp for MSRingWriter" << std::endl;
	exit(1);
    }
//::    std::cout << "debug: MSRingWriter created" << std::endl;
}

MSRingWriter::~MSRingWriter()
{
    if (ring_conn) {
	dl_disconnect (ring_conn);
	dl_freedlcp (ring_conn);
    }
    if (ringserver) free (ringserver);
    if (client_id) free (client_id);
//::    std::cout << "debug: MSRingWriter destroyed" << std::endl;
}

int  MSRingWriter::InitializeRing()
{
    ack_flag = 0;
    /* Loop until connection is made to the ringserver. */
    while (dl_connect(ring_conn) < 0) {
	sleep (1);
    }
    return(TN_SUCCESS);
//::    std::cout << "debug: ring opened: fd = " << ringfd << std::endl;
}

int MSRingWriter::SendMsg(const char* msg,
			 const int   nbytes,
			 const char *station,
			 const char *network,
			 const char *channel,
			 const char *location,
			 const double starttime)
{
    int64_t pktid;
    char streamid[80];
    DATA_HDR *hdr;
    EXT_TIME st, et;
    dltime_t dlstart, dlend;

    // Construct and Ringserver packet.

    if (nbytes > MAX_MSRECORD_SIZE) {
//::	std::cout << "debug: nbytes " << nbytes << " > MAX_MSRECORD_SIZE" << std::endl;
	return (TN_FAILURE);
    }
    hdr = decode_hdr_sdr ((SDR_HDR *)msg, nbytes);
    if (hdr) {
        trim (hdr->station_id);
        trim (hdr->network_id);
        trim (hdr->channel_id);
        trim (hdr->location_id);
    }
    else {
//::	std::cout << "debug: decode_hdr_sdr failed" << std::endl;
        return (TN_FAILURE);
    }

    st = int_to_ext (hdr->begtime);
    et = int_to_ext (hdr->endtime);
    dlstart = dl_time2dltime(st.year, st.doy, st.hour, st.minute, st.second, st.usec);
    dlend = dl_time2dltime(et.year, et.doy, et.hour, et.minute, et.second, et.usec);
    sprintf( streamid, "%s_%s_%s_%s/MSEED", hdr->network_id,
            hdr->station_id, hdr->location_id, hdr->channel_id);

    free_data_hdr (hdr);

    // Write to ringserver.  Return success or failure to caller.
    pktid = dl_write (ring_conn, (void *)msg, nbytes, streamid, dlstart, dlend, ack_flag);
    if (pktid < 0) return(TN_FAILURE);

//:: Use the following code if you want to try to reconnect to the ringserver.
//::    while ((pktid = dl_write (ring_conn, (void *)msg, nbytes, streamid, dlstart, dlend, ack_flag)) < 0) {
//::	// Error writing to ring.  Try to close and reconnect.
//::	std::cout << "Attempt to reconnect to ringserver" << std::endl;
//::	sleep (1);
//::	dl_reconnect(ring_conn);
//::    }

    // std::cout << "debug: ringput returned " << res << std::endl;
    return(TN_SUCCESS);
}

int MSRingWriter::RingIsOpen()
{
    if(ring_conn && ring_conn->link >= 0) {
	return(TN_TRUE);
    }
    else {
	return(TN_FALSE);
    }
}

MSRingWriter& MSRingWriter::operator=(const MSRingWriter &c)
{
    this->ring_conn = dl_newdlcp((char *)"",(char *)"");
    void *p = this->ring_conn;
    memcpy (p, (void *)c.ring_conn, sizeof(struct DLCP_s));
    this->ringserver = strdup(c.ringserver);
    this->client_id = strdup(c.client_id);
    this->ack_flag = c.ack_flag;
    return(*this);
}
