/***********************************************************

File Name : 
	netchancfg.C


Programmer:
	Phil Maechling, adapted by Doug Neuhauser

Description:
	Scan Network, Station, Channel configuration file.
	This accepts a configuration
	file name and a array of channel structure, and it loads the
	array with data from the configuration file.

Creation Date:
	8 August 1996, adapted 03 May 2000


Usage Notes:



Modification History:

16 August 2004 - Added checks to prevent chans array overflow.
		 Added reading of input and output location codes.
                 Added config directive to map input stat, net, chan, loc
		 into corresponding output fields.

**********************************************************/
#include <iostream>
#include <stdio.h>
#include <string.h>
#include "seismic.h"
#include "RetCodes.h"
#include "netchancfg.h"
#include "netchancfg.h"

int scan_netchannel_config_file(char* cfg_file,
                         struct channel_identifier_type chans[],
			 int& number_of_channels_found,
			 int debug_flag)
{
  FILE *cfgfp;
  cfgfp = fopen(cfg_file,"r");
  if (cfgfp == NULL)
  {

    std::cout << "Error opening the configuration file" << std::endl;
    std::cout << "Failed opening " << cfg_file << std::endl;
    return(TN_FAILURE);
  }

  char inputline[MAX_CHARS_IN_FILE_LINE];

  // Define scanned values

  int res;
  char* cres;
 
  number_of_channels_found = 0;

  while ( !feof(cfgfp) )
  {

    cres = fgets(inputline,sizeof(inputline),cfgfp);

    if (feof(cfgfp))
    {
      if(0)
      {
        std::cout << "Found end of channel.cfg configuration file" << std::endl;
      }
      fclose(cfgfp);
      return(TN_SUCCESS);
    }    

    if (cres == NULL)
    {
      std::cout << "Error scanning channel.cfg configuration file" << std::endl;
      fclose(cfgfp);
      return(TN_FAILURE);
    }

    // Check for leading # 
    
    res = strncmp("#",inputline,1);
    if(res == 0)
    {
      // Found comment line
      continue;
    } 
   
    if(number_of_channels_found == MAX_CHANNELS_IN_NETWORK) 
    {
      std::cout << "Maximum channels (" << MAX_CHANNELS_IN_NETWORK << ") allowed reached" <<std::endl;
      std::cout << "Program must be recompiled for more channels in config file"<<std::endl;
      return(TN_FAILURE);
    }

    char network[MAX_CHARS_IN_NETWORK_STRING];
    char station[MAX_CHARS_IN_STATION_STRING];
    char channel[MAX_CHARS_IN_CHANNEL_STRING];
    char location[MAX_CHARS_IN_LOCATION_STRING];
    char seednetwork[MAX_CHARS_IN_NETWORK_STRING];
    char seedstation[MAX_CHARS_IN_STATION_STRING];
    char seedchannel[MAX_CHARS_IN_CHANNEL_STRING];
    char seedlocation[MAX_CHARS_IN_LOCATION_STRING];

    // Save a copy of the line for error reporting sscanf appears to mangle it.
    char saveline[MAX_CHARS_IN_FILE_LINE];
    strcpy (saveline, inputline);    

    res = sscanf(inputline,"%[^.].%[^.].%[^.].%s %[^.].%[^.].%[^.].%s",
	         station,network,channel,location,
		 seedstation,seednetwork,seedchannel,seedlocation);
    
    if (res != 8)
    {
      std::cout << "Error scanning configuration file, line = " << saveline << std::endl;
      return(TN_FAILURE);
    }

    // Convert "-" to blank for location field.
    if (location[0] == '-') location[0] = '\0';
    if (seedlocation[0] == '-') seedlocation[0] = '\0';

    // Move to calling structure.

    strcpy(chans[number_of_channels_found].network,network);
    strcpy(chans[number_of_channels_found].station,station);
    strcpy(chans[number_of_channels_found].channel,channel);
    strcpy(chans[number_of_channels_found].location,location);

    /* `*' in an output field means we use the same value as
       the corresponding input field. */
    if (seednetwork[0] == '*')
	strcpy(chans[number_of_channels_found].seednetwork,network);
    else
	strcpy(chans[number_of_channels_found].seednetwork,seednetwork);
    if (seedstation[0] == '*')
	strcpy(chans[number_of_channels_found].seedstation,station);
    else
	strcpy(chans[number_of_channels_found].seedstation,seedstation);
    if (seedchannel[0] == '*')	
	strcpy(chans[number_of_channels_found].seedchannel,channel);
    else
	strcpy(chans[number_of_channels_found].seedchannel,seedchannel);
    if (seedlocation[0] == '*')
	strcpy(chans[number_of_channels_found].seedlocation,location);
    else
	strcpy(chans[number_of_channels_found].seedlocation,seedlocation);

    if(debug_flag)
    {
      std::cout << "SNCL Input=" << station << "." << network << "." 
	   << channel << "." << location;
      std::cout << " Output=" << seedstation << "." << seednetwork << "."
	   << seedchannel << "." << seedlocation << std::endl;
    } 

    ++number_of_channels_found;

  } // end reading configuration file

  return(TN_SUCCESS);

}
