/***********************************************************

File Name :
        pressor.C 

Programmer:
        Phil Maechling

Description:
        Compressor
	This is a waveform compression routine. It takes an tracebuf
	header, and an array of samples, and compresses them into
	a series of MSEED packets.

Creation Date:
        13 March 1999


Usage Notes:
	There is no attempt to dynamically buffer the samples and to
	generate "always full" MSEED packets. This takes a static
	number of samples, and compresses what it has. So this
	can generate nearly empty SEED Packets.

	This routine also does a lot of dynamic allocation. It malloc
	memory on every packet. 

	Because of this, I recommend running it on a machine with steady
	resource usage, so that the resource don't become limited during
	the big event, and suddenly the mallocing doesn't work just
	when you need it most.

Modification History:

16 August 2004 - Added support for earthworm location codes

31 December 2007 - leap second changes with tntime classes

**********************************************************/
// Various include files
#include <cstdio>
#include <ctime>
#include <climits>
#include <cmath>
#include <cstdlib>
#ifdef __SUNPRO_CC
#include <math.h>
#endif
#include "GenLimits.h"
#include "RetCodes.h"
#include "pressor.h"
#include "seismic.h"

const int pressor_debug =  TN_FALSE;

Compressor::Compressor()
{
  p_first_header_in_session = TN_TRUE;
  p_sequence_number = 0;

  if(get_my_wordorder()==MS_ERROR){
    std::cout << "FATAL Error (ew2msring): Error getting Wordorder via qlib2"<< std::endl;
    exit(1);
  }
}

Compressor::~Compressor()
{


}

int  Compressor::MakePackets(int* counts,
                             int* diffs,
			     const TRACE2_HEADER* trace_hdr,
			     const int nsamples,	
                             const int data_format,
			     const Channel *seedchannel,
			     MSRingWriter& ring)
{

  double samprate;
  int retval;
  char *msbuf;
  char **msbufptr;
  int numblocks;
  int numsamples;
  char errmsg[MAXSTR];
  struct timeval t;
  INT_TIME ttime;
  DATA_HDR dh, *hdr;
  char  msg[BYTES_IN_SEED_TELEMETRY_BLOCK];

  int* t_counts = counts;
  int* t_diffs = diffs;
  
  samprate = trace_hdr->samprate;
  msbufptr = &msbuf;
  errmsg[0] = '\0';

  *msbufptr = NULL;

  // Initialize the data header structure on first time through

  if(p_first_header_in_session)
  {
    init_data_hdr(&p_header);
    p_first_header_in_session = TN_FALSE;
  }

  delete_blockette(&p_header, -1);
  p_header.blksize = BYTES_IN_SEED_TELEMETRY_BLOCK;
  p_header.data_type = data_format;

  memset(p_header.station_id, 0, DH_STATION_LEN + 1);
  memset(p_header.network_id, 0, DH_NETWORK_LEN + 1);
  memset(p_header.channel_id, 0, DH_CHANNEL_LEN + 1);
  memset(p_header.location_id, 0, DH_LOCATION_LEN + 1);
  strncpy(p_header.station_id, seedchannel->station, DH_STATION_LEN);
  strncpy(p_header.network_id, seedchannel->network, DH_NETWORK_LEN);
  strncpy(p_header.channel_id, seedchannel->channel, DH_CHANNEL_LEN);
  strncpy(p_header.location_id, seedchannel->location, DH_LOCATION_LEN);

  p_header.sample_rate = lround(samprate);
  p_header.num_samples = trace_hdr->nsamp;
  p_header.seq_no = p_sequence_number;

  ttime = nepoch_to_int(trace_hdr->starttime);

  p_header.begtime = ttime;
  p_header.hdrtime = ttime;

  if(add_required_miniseed_blockettes(&p_header)<0) {
    std::cout << "Error (ew2msring): building MiniSEED records in add_required_miniseed_blockettes() - " 
	     << errmsg << std::endl;
    return(TN_FAILURE);
  }

  int nleft = p_header.num_samples;
  while (nleft > 0) {
    retval = ms_pack2_steim(&p_header, NULL, t_counts, t_diffs, nleft,
			    &numblocks, &numsamples, msbufptr, 0, errmsg);

    // Log any errors packing MiniSEED data.
    if (retval != 0 && retval != MS_COMPRESS_ERROR) {
      std::cout << "Error (ew2msring): Error packing MiniSEED records - " 
		<< errmsg << std::endl;
      if (*msbufptr != NULL) {
	free(*msbufptr);
      }
      return(TN_FAILURE);
    }

    else {
      if (*msbufptr == NULL && numblocks > 0) {
	  std::cout << "Error (ew2msring): MiniSEED buffer is NULL" << std::endl;
	  return(TN_FAILURE);
      }

    if(pressor_debug) {
        std::cout << "Number of blocks " << numblocks << " from samples " <<
	    numsamples << " to ring" << std::endl;
      }

      // Output records that have been packed.
      for(int i = 0;i<numblocks;i++) {
	int sendval;
	memcpy(msg,(*msbufptr)+(i*BYTES_IN_SEED_TELEMETRY_BLOCK),
				BYTES_IN_SEED_TELEMETRY_BLOCK);
	hdr = decode_hdr_sdr ((SDR_HDR *)msg, BYTES_IN_SEED_TELEMETRY_BLOCK);
	double starttime = (double)unix_time_from_int_time(hdr->begtime) +
	    ((double)(hdr->begtime.usec)/USECS_PER_SEC);
	free_data_hdr (hdr);

	sendval = ring.SendMsg(msg,BYTES_IN_SEED_TELEMETRY_BLOCK,
			     seedchannel->network,
			     seedchannel->station,
			     seedchannel->channel,
			     seedchannel->location, 
			     starttime);

        if (pressor_debug) {
	  std::cout << "Sending packet for " << seedchannel->station << "." 
					     << seedchannel->network << "." 
					     << seedchannel->channel << "." 
					     << seedchannel->location 
					     << " for time " << std::fixed << starttime 
					     << " nsamples " << hdr->num_samples
					     << " success=" << sendval << std::endl;
	}

        if (sendval != TN_SUCCESS) {
          std::cout << "Error (ew2msring): Writing MiniSEED records to ring" << std::endl;
	  free(*msbufptr);
	  return(TN_FAILURE);
	}
        // Update the sequence number.  The sequence number is now ready for next call.
	p_sequence_number = (p_sequence_number + numblocks) % 1000000;
      }
 
      // Skip a few raw samples if we had a compression error creating the MiniSEED.
      int nskip = 0;
      if (retval == MS_COMPRESS_ERROR) {
	int nskip = (nleft > 5) ? 5 : nleft;
	std::cout << "Compressed " << numsamples << " out of " << nleft << " - Skipping " << nskip << " samples." << std::endl;
        p_header.xm2 = p_header.xm1 = 0;
        numsamples += nskip;
      }
      // If we still have samples to compress, compute new times and new pointers for raw counts and diffs.
      if (nsamples < nleft) {
	int sec, usec;
	time_interval2 (nsamples, p_header.sample_rate, p_header.sample_rate_mult, &sec, &usec);
	p_header.hdrtime = add_time (p_header.hdrtime, sec, usec);
	p_header.begtime = add_time (p_header.begtime, sec, usec);
	t_counts += nsamples;
	t_diffs += nsamples;
      }

      nleft -= numsamples;
      if (*msbufptr) free(*msbufptr);
      *msbufptr = NULL;
    }
  }

  return(TN_SUCCESS);
}
